# wx_square_loop_colors_snap_tracking.py  24Jan2025  crs, track
# Display a square with colored sides

import graphics_braille.wx_turtle_braille as tu    # Get our graphics

from graphics_braille.select_trace import SlTrace
import graphics_braille.canvas_copy as canvas_copy
SlTrace.setFlags("snapshot")
SlTrace.setFlags("user,host,rpc")
colors = ["red","orange","yellow","green"]
canvas =   tu.getcanvas()
cv_str = canvas_copy.canvas_show_items(canvas,
                    show_coords=False,
                    show_options=True,
                    use_value_cache=True)

SlTrace.lg(f"Beginning: cv_str:{cv_str}", "canvas_copy")
nside = 0
tu.snapshot(f"Begining - before square")
for color in colors:
    nside += 1
    tu.width(40)
    tu.color(color)
    tu.forward(200)
    tu.right(90)
    cv_str = canvas_copy.canvas_show_items(canvas,
                    show_coords=False,
                    show_options=True,
                    use_value_cache=True)
    SlTrace.lg(f"side{nside}: {color}: cv_str:{cv_str}", "canvas_copy")
    tu.snapshot(f"side{nside}: {color}")
tu.done()		    # Complete drawings



# wx_square_loop_colors.py  26Jul2024  crs, use wx_turtle_braille.py directly
# Display a square with colored sides

import graphics_braille.wx_turtle_braille as tu    # Get our graphics
##from turtle import *		     # Get standard stuff
from graphics_braille.select_trace import SlTrace
import graphics_braille.canvas_copy as canvas_copy

#SlTrace.setFlags("user,host")
colors = ["red","orange","yellow","green"]

nside = 0
#snapshot(f"Begining - before square")   # WHY to get 1'st side
for color in colors:
    nside += 1
    SlTrace.lg(f"before side {nside}")
    tu.width(40)
    tu.color(color)
    tu.forward(200)
    tu.right(90)
    tu.snapshot(f"side{nside}: {color}")
tu.done()		    # Complete drawings



"""
Created on Oct 6, 2018

@author: raysm
"""
from tkinter import *

class SelectDDChoice(object):
    """
    Dropdown choice one of n text strings
    """


    def __init__(self, ctl_frame,
                    selection,
                    default=None):
        """ Setup Dropdown menu for selection
        :ctl_frame: containing / controling frame/widget
        :selection:  list  of text strings
        :default: default text string - None  - use first
        """
        tkvar = StringVar(ctl_frame)

        if default is None:
            default = selection[0]
        tkvar.set(default)
        self.tkvar = tkvar
        sel_dict = {}
        for s in selection:
            sel_dict[s] = s
        popup_menu = OptionMenu(ctl_frame,  tkvar, *sel_dict)
        self.popup_menu = popup_menu
        ###popup_menu.pack()
        
        
    def get(self):
        return self.tkvar.get()


    def pack(self, **kwargs):
        self.popup_menu.pack(kwargs)
        
        
    def set(self, val):
        self.tkvar.set(val)

    
    def set_field(self, val):
        self.set(val)
        
        
            
if __name__ == '__main__':
    root = Tk()
    root.title("SelectDDChoice test")
    
    ctl_frame = Frame(root)
    ctl_frame.pack()
    sel=["random", "ascend", "descend"]
    sel_default="random"
    val_entry = SelectDDChoice(ctl_frame,
                                selection=sel,
                                default=sel_default)
    value_type = str
    # on change dropdown value
    def change_dropdown(*args):
        print( val_entry.tkvar.get() )
     
    # link function to change dropdown
    val_entry.tkvar.trace('w', change_dropdown)
 
    root.mainloop()


#__init__.py    30Oct2025  crs
# Command line scripts for graphics_braille
import os
import shutil
import pathlib
import importlib.resources

import graphics_braille
from graphics_braille.onedrive_ck import is_desktop_on_onedrive

def main_function():
    print("Installation Run")

def gbhelp():
    print("""
          gb, gbhelp - List graphics-braille commands
          gbreadme - Display README.md
          gbfiles - add user test files to user desktop
          gbtest - Do simple program test
          gbtest2 - Do simple program test - spokes
          """)

def gbreadme():
    import os
    os.system("distinfo_readme graphics_braille")

def gbfiles():
    """ Place/Update user files on Desktop
    """
    package_name = "graphics_braille"
    package_traversable = importlib.resources.files(package_name)
    print(f"{package_traversable = }")
    data_dir_base = "gb_exercises"
    data_path = package_traversable.parent / data_dir_base
    print(f"Path to data directory: {data_path}")    
    source_dir = str(data_path)
    print(f"{source_dir =}")                                    
    desktop = pathlib.Path.home() / 'Desktop'
    on_onedrive, desktop_location = is_desktop_on_onedrive()
    if on_onedrive:
        print(f"OneDrive {desktop_location = }")
        desktop = desktop_location
    try:
        # Copy the directory tree, allowing existing directories
        src = source_dir
        dst = os.path.join(desktop, os.path.basename(src))
        os.makedirs(dst, exist_ok=True) # Ensure the parent exists if needed
        shutil.copytree(src, dst, dirs_exist_ok=True)
        print(f"Successfully copied '{src}' to '{dst}'.")
    except Exception as e:
        print(f"An error occurred: {e}"
              f" in copy '{src}' to '{dst}'.")
    
    
    
def gbtest():
    import graphics_braille.z_show_square_loop_colors_braille
    
def gbtest2():
    import graphics_braille.show_spokes_braille



# select_corner.py        

from math import sqrt

from graphics_braille.select_trace import SlTrace
from graphics_braille.select_error import SelectError
from graphics_braille.select_part import SelectPart

class SelectCorner(SelectPart):
    corner_width_display = 8    # Default display size of corner in pixels
    corner_width_select = 8    # Default select size of corner in pixels
    corner_width_standoff = 10    # Default standoff size of corner in pixels
    corner_fill = "red" # Default corner color
    corner_fill_highlight = "pink"      # Default corner highlight color


    '''Use default copy.deepcopy()
    def __deepcopy__(self, memo=None):
        """ provide deep copy as a custimized "constructor",
        reducing recusion
        """
        res = self.copy()        
        return res
    '''

    
    def __init__(self, sel_area, **kwargs):
        super().__init__(sel_area, "corner", **kwargs)
                    
            
    def display(self):
        """ Display corner on inside upper left corner
        """
        SlTrace.lg("%s: %s at %s" % (self.part_type, self, str(self.loc)), "display")
        
        if not self.sel_area.display_game:
            return

        self.display_clear()
        if self.is_highlighted():
            """ Highlight given corner
            :hand: Corner handle
            :Returns: object tag for deletion
            """
            c1x,c1y,c3x,c3y = self.get_rect(enlarge=True)
            if self.display_shape == "circle":
                self.highlight_tag = self.sel_area.canvas.create_oval(
                                c1x, c1y, c3x, c3y,
                                fill=SelectPart.corner_fill_highlight)
            else:
                self.highlight_tag = self.sel_area.canvas.create_rectangle(
                                c1x, c1y, c3x, c3y,
                                fill=SelectPart.corner_fill_highlight)
        else:
            loc = self.loc 
            SlTrace.lg("corner: %s" % str(loc), "display")
            c1x,c1y,c3x,c3y = self.get_rect()
            if self.display_size is not None:
                cwidth = self.display_size
                cheight = cwidth
                cx = (c3x-c1x)/2 + c1x 
                cy = (c3y-c1y)/2 + c1y
                c1x = cx - cwidth/2
                c3x = cx + cwidth/2
                c1y = cy - cheight/2
                c3y = cy + cheight/2
            if (self.display_shape is not None
                 or self.display_size is not None):
                if self.display_shape == "circle":
                    tags = []
                    tag = self.sel_area.canvas.create_oval(
                        c1x, c1y, c3x, c3y, fill=self.color)
                    tags.append(tag)
                    cc_x, cc_y, cc_sx, cc_sy = self.get_center_size()
                    SlTrace.lg("cc_x,y=%d,%d  cc_sx,y=%d,%d" % (cc_x,cc_y, cc_sx,cc_sy), "part_info")
                    tag = self.sel_area.canvas.create_line(cc_x,cc_y, cc_x,cc_y-cc_sy,
                                                        cc_x,cc_y, cc_x-cc_sx, cc_y,
                                                        cc_x,cc_y, cc_x+cc_sx, cc_y,
                                                        cc_x,cc_y, cc_x, cc_y+cc_sy,
                                                        cc_x,cc_y, cc_x, cc_y-cc_sy,
                                                          )
                    tags.append(tag)
                    self.display_tag = tags
                else:
                    tags = []
                    tag = self.sel_area.canvas.create_rectangle(
                                c1x, c1y, c3x, c3y, fill=self.color)
                    tags.append(tag)
                    cc_x, cc_y, cc_sx, cc_sy = self.get_center_size()
                    tag = self.sel_area.canvas.create_line(cc_x,cc_y, cc_x,cc_y-cc_sy,
                                                        cc_x,cc_y, cc_x-cc_sx, cc_y,
                                                        cc_x,cc_y, cc_x+cc_sx, cc_y,
                                                        cc_x,cc_y, cc_x, cc_y+cc_sy,
                                                        cc_x,cc_y, cc_x, cc_y-cc_sy,
                                                          )
                tags.append(tag)
                self.display_tag = tags
            else:
                self.display_tag = self.sel_area.canvas.create_rectangle(
                        c1x, c1y, c3x, c3y, fill=self.color)


    def distance(self, cx, cy):
        """ Distance between corner and coordinates cx,cy
        :cx: x coordinate
        :cy: y coordinate
        """
        co_x,co_y,co_xsize,co_ysize = self.get_center_size()
        return sqrt((co_x-cx)**2+(co_y-cy)**2)


    def get_nodes(self, indexes=None):
        """ Find nodes/points of part
        return pairs (index, node)
        """
        return [(0,self.loc.coord)]


    def set_node(self, index, node):
        """ Set node to new value
        """
        if index != 0:
            raise SelectError("set_node %s non-zero index %d"
                              % (self, index))
        self.loc.coord = node


    def edge_dxy(self):
        """ Get "edge direction" as x-increment, y-increment pair
        """
        return 0,0                  # No change
        
                     

    
    def get_rect(self, sz_type=None, enlarge=False):
        """ Get rectangle containing edge handle
        Use connected corners
        Coordinates returned are ordered ulx, uly, lrx,lry so ulx<=lrx, uly<=lry
        We leave room for the corners at each end
        :edge - selected edge
        :enlarge - True - enlarge rectangle
        """
        wlen = self.get_edge_width(sz_type=sz_type, enlarge=enlarge)
        
        c1x = self.loc.coord[0]
        c1y = self.loc.coord[1]
        c3x = c1x
        c3y = c1y
        if c1x >= wlen/2:     # Yes - widen the orthogonal dimension
            c1x -= wlen/2
        c3x += wlen
        if c1y >= wlen:     # Yes - widen the orthogonal dimension
            c1y -= wlen/2
        c3y += wlen
        if enlarge:
            wenlarge = self.edge_width_enlarge
            c1x -= wenlarge
            c1y -= wenlarge
            c3x += wenlarge
            c3y += wenlarge
                
        return c1x,c1y,c3x,c3y
    
    def highlight(self):
        """ Highlight and display
        """
        self.highlight_set()
        self.display()


    
    def loc_key(self):
        """ Return location of part as a key
        """
        key = self.loc.coord
        return (key)



# s_show_square_loop_colors_w80_braille.py  # braille
# direct e.g. import xxxxxx as x, then use as x.mod
# Display a square with colored sides

##from turtle import *		     # Get standard stuff
import graphics_braille.wx_turtle_braille as tu		     # Get braille

colors = ["red","orange","yellow","green"]

for colr in colors:
    tu.width(80)
    tu.color(colr)
    tu.forward(200)
    tu.right(90)
tu.done()		    # Complete drawings



# select_list_ckbox.py 14Apr2020    crs
"""
TEMPORARY - to replace select_list with ckbox option
Select list choose from a list of strings - returning selected
Adapted from traceControlWindow
"""
from tkinter import *
import atexit

import time

from select_trace import SlTrace
from select_error import SelectError
from image_hash import ImageHash

class SelectList(Toplevel):
    def __init__(self, tcbase=None, title=None, items=None,
                 item_sep=None,
                 ckbox=False,
                 entry_text=None,
                 position=None,
                 size=None,
                 image_size=None, image_hash=None, default_to_files=False, cancel_value=None):
        """ 
        :tcbase: - parent - call basis must have tc_destroy to be called if we close
        :title: Window title
        :items: list of strings / or file names if (<file:....>
            files are displayed as images
        :item_sep: if present, add in separator line where items list contains
                matching this matching text
        :ckbox: True - provide check boxes for all entries
                returns: list of checked
        :position: (x,y) in pixels on screen
        :size: (width, hight) in in pixels of window
        :entry_text: If present, include entry at top of list
                    Place text as initial entry contents
        :image_size: (width, height) of image in pixels
        :image_hash: access to image files
        :default_to_files: True = default to file names else assume text entry fields
                        default: false
        """
        self.done = False           # Set True if done (OK, Cancel)
        self.ok = False             # Set if OK
        self.canceled = False
        self.standalone = False      # Set True if standalone operation
        self.ckbox = ckbox
        if ckbox:
            self.ckbox_vars_by_item = {}    # corresponds 1-to-1 self.items
        self.items = items
        self.item_sep = item_sep
        if image_size is None:
            image_size = (125, 125)
        self.image_size = image_size
        self.default_to_files = default_to_files
        if tcbase is None:
            SlTrace.lg("Standalone SelectList")
            tcbase = Toplevel()
            frame = Frame(tcbase)
            frame.pack()
            self.standalone = True
            ###tcbase.withdraw()
        self.tc_mw = self.tcbase = tcbase
        if position is None:
            position = (800, 100)
        if size is None:
            size = (200, 200)
        x0 = position[0]
        y0 = position[1]
        w = size[0]
        h = size[1]
        tc_geo = f"{w}x{h}+{x0}+{y0}"
        self.tc_mw.geometry(tc_geo)
        if title is None:
            title = "Testing"
        self.entry_text = entry_text
        self.entry_text_var = None      # Text Entry
        self.tc_mw.title(title)
        top_frame = Frame(self.tc_mw)
        top_frame.pack(side="top", fill="both", expand=False)
        self.top_frame = top_frame
        if title is not None:
            title_label = Label(master=self.top_frame, text=title, font="bold")
            title_label.pack(side="top", fill="both", expand=False)
        ok_button = Button(master=self.top_frame, text="OK", command=self.on_ok)
        ok_button.pack(side="left", fill="both", expand=False)
        cancel_button = Button(master=self.top_frame, text="CANCEL", command=self.on_cancel)
        cancel_button.pack(side="right", fill="both", expand=False)
        self.tc_mw.protocol("WM_DELETE_WINDOW", self.delete)
        top_frame.pack(side="top", fill="both", expand=False)
        tc_frame = Frame(self.tc_mw)
        tc_frame.pack(side="top", fill="both", expand=True)
        self.tc_frame = tc_frame
        self.tc_text_frame = None           # So we can destroy / resetup in create_items_region
        self.items_text = None
        self.sb = None
        if image_hash is None:
            image_hash = ImageHash()
        self.image_hash = image_hash
        self.selected_text_field = None
        self.cancel_value = cancel_value
            
        self.create_items_region(items=items)
            
    def create_items_region(self, items=None):
        if self.tc_text_frame is not None:
            self.tc_text_frame.pack_forget()
            self.tc_text_frame.destroy()
            self.tc_text_frame = None
            
        if self.sb is not None:
            self.sb.destroy()
        self.update()                           # Show progress
        self.tc_text_frame = Frame(self.tc_frame)
        self.tc_text_frame.pack(side="top", fill="both", expand=True)

        self.start = 0
        self.sb = Scrollbar(master=self.tc_text_frame, orient="vertical")
        max_width = 5
        min_height = 10
        t_height = min_height
        max_height = 20
        nfound = 0
        n_text = 0
        for item in items:
            if self.item_sep is not None and item == self.item_sep:
                continue
            if self.is_image(item):
                width = 10
            else:
                n_text += 1             # Count text entries
                width = len(item)
            if width > max_width:
                max_width = width
            nfound += 1
        win_width = max_width
        if nfound < min_height:
            t_height = min_height
        if nfound > max_height:
            t_height = max_height
            
        text_region = Text(self.tc_text_frame, width=win_width, height=t_height,
                    yscrollcommand=self.sb.set,
                    state=DISABLED)
        self.sb.config(command=text_region.yview)
        self.sb.pack(side="right",fill="y")
        text_region.pack(side="top", fill="both", expand=True)
        self.update()                           # Show progress
        self.data_by_widget = {}      # Dictionary by widget to  (btn, text_field, image)
        if self.entry_text is not None:
            max_width = max(int(self.image_size[0]/6), max_width)
            self.add_text_entry(text_region, text=self.entry_text, width=max_width)                
        for item in items:
            if self.item_sep is not None and item == self.item_sep:
                self.add_separator(text_region)
                continue
            if self.ckbox:
                if self.is_image(item):
                    image = self.get_image(self.image_name(item), width=max_width)
                    self.add_image_ckbox(text_region, image=image, text_field=item)
                else:
                    self.add_text_ckbox(text_region, item, width=max_width+3)                
            else:
                if self.is_image(item):
                    image = self.get_image(self.image_name(item), width=max_width)
                    self.add_image_button(text_region, image=image, text_field=item)
                else:
                    self.add_text_button(text_region, item, width=max_width+3)                
        self.update()                           # Show progress
        if self.standalone:
            atexit.register(self.on_exit)
            self.update_loop()

    def add_image_button(self, text_region, image=None, text_field=None):
        
        btn = Button(text_region, image=image, command=lambda : self.do_image_button(text_field))
        self.data_by_widget[btn] = (btn, text_field, image)     # field is what we need
        text_region.config(state=NORMAL)
        text_region.window_create("end", window=btn)
        text_region.insert("end", "\n")

    def add_separator(self, text_region):
        """ Add check box entry
        """
        text_region.config(state=NORMAL)
        text_region.insert("end", "\n")
        text_region.config(state=DISABLED)

    def add_text_button(self, text_region, text=None, width=None):
        if text is None:
            text = ""
        if width is None:
            width = len(text) + 3
        btn = Button(text_region, text=text, command=lambda : self.do_text_button(text))
        self.data_by_widget[btn] = (btn, text)     # field is what we need
        text_region.config(state=NORMAL)
        text_region.window_create("end", window=btn)
        text_region.insert("end", "\n")

    def add_text_ckbox(self, text_region, text=None, width=None):
        """ Add check box entry
        """
        if text is None:
            text = ""
        if width is None:
            width = len(text) + 3
        ckbox_var = BooleanVar()
        ckbox = Checkbutton(text_region, text=text, variable=ckbox_var)
        self.data_by_widget[ckbox] = (ckbox, ckbox_var, text)     # field is what we need
        self.ckbox_vars_by_item[text] = ckbox_var
        text_region.config(state=NORMAL)
        text_region.window_create("end", window=ckbox)
        text_region.insert("end", "\n")
        text_region.config(state=DISABLED)

    def add_image_ckbox(self, text_region, image=None, text_field=None):
        SlTrace.lg("add_image_ckbox - TBD")
        
    def add_text_entry(self, text_region, text=None, width=None):
        """ Add text entry, records last entry added"""
        if text is None:
            text = ""
        entry_text_var = StringVar()
        entry_text_var.set(text)
        ent = Entry(text_region, width=width, textvariable=entry_text_var, bd=5, relief=RIDGE)
        self.data_by_widget[ent] = (ent, entry_text_var, text)
        text_region.config(state=NORMAL)
        text_region.window_create("end", window=ent)
        text_region.insert("end", "\n")
        text_region.config(state=DISABLED)
        ent.bind("<Return>", self.enter_entry)
        self.entry_text_var = entry_text_var       # Save for use on OK

    def on_ok(self):
        """ OK button pressed
        """
        SlTrace.lg("OK button pressed")
        if self.ckbox:
            self.ok = True
            self.done = True
            return
        
        if self.entry_text_var is not None:
            text = self.entry_text_var.get()
            self.ok_text(text)
            self.ok = True
        else:
            self.on_cancel()
        
    def on_cancel(self):
        """ Cancel button pressed
        """
        self.delete()
        
    def delete(self):
        """ Called to delete window, e.g. when x was clicked
        """
        self.canceled = True
        self.done = True
        
        
    def do_image_button(self, text_field):
        self.selected_text_field = text_field
        self.ok = True
        self.done = True
        
    def do_text_button(self, text):
        self.selected_text_field = text
        self.ok = True
        self.done = True
        
    def image_name(self, item_text):
        """ Convert item_text to image name
        :item_text: text
        """
        return self.image_hash.image_name(item_text, default_to_files=self.default_to_files)
        
        
    def is_image(self, item_text):
        return self.image_hash.is_file(item_text, default_to_files=self.default_to_files)
            
    def get_image(self, item_text, width=None):
        image_size = self.image_size
        image = self.image_hash.get_image(item_text, size=image_size)
        return image
        
    def get_selected(self, return_text=False):
        """ Return selected field: <file:....> for selected image file, else text entry
        :give_text: return text value
                default: return image string, if image
        """
        if self.ckbox:
            raise SelectError("Use get_checked, NOT get_selected for ckbox")
        
        while not self.done:                
            self.sleep(.1)
        self.delete_tc_window()
        if self.canceled:
            text_field = self.cancel_value
        else:
            text_field = self.selected_text_field
        if return_text:
            return text_field
        
        if self.is_image(text_field):
            text_field = self.image_hash.name2image_string(text_field)
        return text_field
        
    def get_checked(self):
        """ Return list of checked: <file:....> for selected image file, else text entry
        """
        if not self.ckbox:
            raise SelectError("Don't use get_checked unless ckbox==True")
        while not self.done:
            self.sleep(.1)
        self.delete_tc_window()
        if self.canceled:
            return []    
                
        if self.ok:
            checked_items = []
            for item in self.items:
                if self.item_sep is not None and item == self.item_sep:
                    continue        # Skip separator
                ckbox_var = self.ckbox_vars_by_item[item]
                cked = ckbox_var.get()
                if cked:
                    checked_items.append(item)
            return checked_items
        
        return []
        
    def on_exit(self):
        """ Close down window on program exit
        """
        SlTrace.lg("Closing down Trace Control Window")
        self.delete_tc_window()
        
    def update_loop(self):
        """ continue repeated tk.update() calls
        to enable window operation
        """
        loop_time = 50          # Loop recall time (msec)
        self.update()
        self.tc_mw.after(loop_time, self.update_loop)

    def update(self):
        if self.tc_mw is not None and self.tc_mw.winfo_exists():
            self.tc_mw.update()
        
    def sleep(self, sec):
        """ "sleep" for a number of sec
        without stoping tkinter stuff
        :sec: number of milliseconds to delay before returning
        """
        if self.tc_mw is None:
            return
        
        self.update()             # Insure at least one update
        now = time.time()
        end_time = now + sec
        while time.time() < end_time:
            if self.tc_mw is None:
                return
            
            self.update()
        return
    
    def mainloop(self):
        self.tc_mw.mainloop()
        
    def delete_tc_window(self):
        """ Process Trace Control window close
        """
        if self.tc_mw is not None:
            self.tc_mw.destroy()
            self.tc_mw = None
        
        if self.tcbase is not None and hasattr(self.tcbase, 'tc_destroy'):
            self.tcbase.tc_destroy()
                
    def enter_entry(self, event):
        widget = event.widget
        _, entry_text_var, _ = self.data_by_widget[widget]
        text = entry_text_var.get()
        self.ok_text(text)

    
    
    def ok_text(self, text):
        self.default_to_files = False       # Assuming text (no modification)
        self.selected_text_field = text
        self.ok = True
        self.done = True
        
if __name__ == '__main__':
    def report_change(item, val, cklist=None):
        SlTrace.lg("changed: %s = %d" % (item, val))
        new_val = SlTrace.getLevel(item)
        SlTrace.lg("New val: %s = %d" % (item, new_val))
        if cklist is not None:
            cklist.list_ckbuttons()
    
    root = Tk()
    SlTrace.set_mw(root)
    ###frame = Frame(root)
    ###frame.pack()
    SlTrace.setProps()
    image_hash = ImageHash(image_dir="../images/miscellaneous")
    image_files = image_hash.get_image_files()
    item_sep = "====="
    text_items = ["ONE","TWO", item_sep, "3", "FOUR", item_sep, "5", "6"]
    x0 = 300
    y0 = 400
    width = 200
    height = 400
    SlTrace.lg(f"x0={x0}, y0={y0}, width={width}, height={height}", "select_list")
    
    SlTrace.lg("Text ckbox with item_sep=")                    
    app = SelectList(ckbox=True,
                     items=text_items, position=(x0, y0), size=(width, height),
                     item_sep=item_sep,
                     title="Text Ckbox")
    selected_field = app.get_checked()
    SlTrace.lg(f"text_items: selected_field:{selected_field}")    
    
    SlTrace.lg("Text Buttons")                    
    app = SelectList(items=text_items, position=(x0, y0), size=(width, height),
                     title="Text Buttons")
    selected_field = app.get_selected()
    SlTrace.lg(f"text_items: selected_field:{selected_field}")    

    SlTrace.lg("Text Buttons + entry")
    app = SelectList(items=text_items, position=(x0, y0), size=(width, height),
                     entry_text="Add your own",
                     title="Text Buttons - with entry")
    selected_field = app.get_selected()
    SlTrace.lg(f"text_items + entry: selected_field:{selected_field}")    

    SlTrace.lg("Image Buttons")
    app = SelectList(items=image_files, image_hash=image_hash, default_to_files=True,
                     title="Image Buttons",
                     position=(x0, y0), size=(width, height))
    selected_field = app.get_selected()
    SlTrace.lg(f"image_image: selected_field:{selected_field}")

    SlTrace.lg("Image Buttons + entry")
    app = SelectList(items=image_files, image_hash=image_hash, default_to_files=True,
                     title="Image Buttons",
                     entry_text="Add your own",
                     position=(x0, y0), size=(width, height))
    selected_field = app.get_selected()
    SlTrace.lg(f"image_image: selected_field:{selected_field}")
        
    SlTrace.lg("End of test")


# wx_spokes.py  16Jul2024  crs use wx_turtle_braille directly
#               20Jan2024, crs from spokes.py
# Display a star with spokes

import graphics_braille.wx_turtle_braille as tu        # Set for wxPython
#import turtle as tu    # Bring in turtle graphic functions
tu.speed("fastest")
colors = ["red","orange","yellow",
          "green","blue","indigo","violet"]
for colr in colors:      # do all colors in list
    tu.color(colr)
    tu.forward(300)
    tu.dot(100)
    tu.backward(300)
    tu.right(360/len(colors))
tu.done()



# select_control_window.py
"""
Base for independent control window
NOTE: for Singleton version see select_control_window_singleton.py
Facilitates
    setting and display of game controls
    persistent storage of values
    window positioning / sizing
    Undo / Re-do of value setting
"""
from tkinter import *
import time

from graphics_braille.crs_funs import str2bool, str2val
from graphics_braille.select_error import SelectError
from graphics_braille.select_trace import SlTrace
from graphics_braille.select_input import SelectInput

    

def content_var_type(val):
    """ Convert content_variable instance
    into variable type
    :var:   variable type
    """
    if isinstance(val, StringVar):
        val_type = str
    elif isinstance(val, IntVar):
        val_type = int
    elif isinstance(val, DoubleVar):
        val_type = float
    elif isinstance(val, BooleanVar):
        val_type = bool
    else:
        raise SelectError(f"Unsupported content var type {val}")
    return val_type
    

def content_var(val):
    """ create content variable of the type of val
    :val:   value to initialize content
    """
    val_type = type(val)
    if val_type == str:
        var = StringVar()
    elif val_type == int:
        var = IntVar()
    elif val_type == float:
        var = DoubleVar()
    elif val_type == bool:
        var = BooleanVar()
    else:
        raise SelectError(f"Unsupported content var val_type {val_type}")
    var.set(val)
    return var
    
    
class SelectControlWindow(Toplevel):
    CONTROL_NAME_PREFIX = "window_control"
    DEF_WIN_X = 500
    DEF_WIN_Y = 300
    _instance = None         # Instance
    instance_no = 0         # Count instances

    @classmethod
    def reset_class(cls):
        """ Reset to gone
        Facilitate regeneration of window
        """
        cls._instance = None
    
    def __init__(self, mw=None,
                play_control=None,
                control_prefix=None,
                title=None,
                display=True,
                set_cmd=None,
                win_x=None,
                win_y=None,
                win_width=None,
                win_height=None,
                enter_command=None,
                 ):
        """ Control attributes
        :play_control:        special for game play
        :control_prefix: prefix for properties file entries
        :title: window title
        :set_cmd: function, if present, to call on button press
        :win_x:  New window x position default: use properties entry
        :win_y:  New window y position default: use properties entry
        :win_width: New window width default: use properties entry
        :win_height: New window height default: use properties entry
        :enter_command: if present, the common entry element keyboard ENTER command
        """
        SelectControlWindow.instance_no += 1
        self.play_control = play_control
        if control_prefix is None:
            control_prefix = self.CONTROL_NAME_PREFIX
        self.control_prefix = control_prefix
        if mw is None:
            mw = Toplevel()
        self.mw = mw
        self.update()           # TFD debugging display
        self.mw.protocol("WM_DELETE_WINDOW", self.delete_window)    
        if title is None:
            title = "Game Control"
        self.title = title
        self.vals = {}          # Current values if any
        self.ctls_labels = {}   # ctl labels (if one) by field
        self.ctls = {}          # Dictionary of field control widgets
        self.ctls_vars = {}     # Dictionary of field control widget variables
        self.display = display   # Done in instance, if at all
        self.set_cmd = set_cmd
        self.enter_command = enter_command
        ndim_spec = 0
        if win_x is not None:
            self.set_prop_val("win_x", win_x)
            ndim_spec += 1
        if win_y is not None:
            self.set_prop_val("win_y", win_y)
            ndim_spec += 1
        if win_width is not None:
            self.set_prop_val("win_width", win_width)
            ndim_spec += 1
        if win_height is not None:
            self.set_prop_val("win_height", win_height)
            ndim_spec += 1
        if ndim_spec > 0:
            self.arrange_windows()
        self._is_displayed = False
        
    def report(self, msg):
        """ Report, via popup window
        :msg:
        """
        SlTrace.report(f"{self.control_prefix}: {msg}")
            
    def set_play_control(self, play_control):
        """ Link ourselves to the display
        """
        self.play_control = play_control
        
            
    def control_display(self):
        """ display /redisplay controls to enable
        entry / modification
        """
        top_frame = Frame(self.mw)
        top_frame.pack(side="top", fill="x", expand=True)
        self.top_frame = top_frame
        
        self.base_frame = top_frame     # Changed on use
        self.base_field = "game_control"
        self.mw.title(self.title)
        top_frame = Frame(self.mw)
        top_frame.pack(side="top", fill="x", expand=True)
        self.top_frame = top_frame
        
        self.title_frame = None     # So can check if title
        if self.title is not None:
            self.title_frame = Frame(top_frame)
            self.title_frame.pack(side="top", fill="x", expand=True)
            title_label = Label(master=self.title_frame, text=self.title, font="bold")
            title_label.pack()
            
        
        bottom_frame = Frame(self.mw, borderwidth=2, relief=SUNKEN)
        bottom_frame.pack(side="bottom", expand=True)
        self.bottom_frame = bottom_frame
        
        self.set_fields(bottom_frame, "base", title="")
        self.set_button(field="set", label="Set", command=self.set)
        self.set_sep()
        self.set_button(field="Reset", label="Reset", command=self.reset)
        self.set_sep()
        self.set_button(field="Undo", label="Undo", command=self.undo)
        self.set_sep()
        self.set_button(field="Redo", label="Redo", command=self.redo)
        
        self.arrange_windows()
        self._is_displayed = True       # Mark as displayed
        

    """ Control functions for game control
    """
    
    def set(self):
        self.set_vals()
        
    
    def reset(self):
        self.set_vals()

    
    def undo(self):
        self.set_vals()

    def sleep(self, sec):
        """ "sleep" for a number of sec
        without stoping tkinter stuff
        :sec: number of milliseconds to delay before returning
        """
        now = time.time()
        end_time = now + sec
        while time.time() < end_time:
            self.update()
        return
    
    def redo(self):
        self.set_vals()

    def update(self):
        """ Do update to see display process
        """
        if self.mw is not None:
            self.mw.update()
            
    def get_ctl(self, field_name):
        """ Get field ctl
        :field_name: field name
        """
        field = field_name.lower()
        if field not in self.ctls_vars:
            fields_str = ", ".join(self.ctls_vars.keys())
            SlTrace.lg(f"ctl_vars fields: {fields_str}")
            raise SelectError(f"get_ctl: {self.control_prefix} has no field {field}")
            
        ctl = self.ctls[field]
        return ctl
            
    def get_val_from_ctl(self, field_name):
        """ Get value from field
        Does not set value
        :field_name: field name
        """
        field = field_name.lower()
        if field not in self.ctls_vars:
            fields_str = ", ".join(self.ctls_vars.keys())
            raise SelectError(f"get_val_from_ctl: '{self.control_prefix}'"
                              f" has no field '{field}'"
                              f"\n\nctl_vars fields: {fields_str}\n")
            
        value = self.ctls_vars[field].get()
        return value

    def update_from_ctl(self, field_name):
        """ Get value from field, and update properties
        :field_name: field name
        :returns: value from field
        """
        val = self.get_val_from_ctl(field_name)
        self.set_prop_val(field_name, val)
        return val
        
    
    def set_vals(self):
        """ Read form, if displayed, and update internal values
        """
        for field in self.ctls_vars:
            self.set_val_from_ctl(field)

    def set_val_from_ctl(self, field_name):
        """ Set ctls value from field
        Also updates player value properties
        :field_name: field name
        """
        if not field_name in self.ctls_vars:
            raise SelectError("No field named %s" % field_name)
        ctl_var = self.ctls_vars[field_name]
        if ctl_var is None:
            raise SelectError("No variable for %s" % field_name)
        while True:
            try:
                value = ctl_var.get()
                break
            except:
                ctl = self.ctls[field_name]
                ctl_str = ctl.get()
                ctl_var_type = content_var_type(ctl_var)
                default_val_str = self.get_prop_val(field_name, None)
                default_val = str2val(default_val_str, ctl_var_type)
                
                sr = SelectInput(master=self, title="GameControl",
                                  message=f"Bad format:'{ctl_str}' for {field_name}: ",
                                  default=default_val)
                value = sr.result
                if value is not None:
                    break
                    
        self.set_ctl_val(field_name, value)        
        self.set_prop_val(field_name, value)


    def set_fields(self, base_frame, base_field, title=None):
        """ Set current control area
        :frame: current frame into which controls go
        :base_field: base for variables/widgets are stored
        """
        base_frame.pack()
        self.base_frame = base_frame
        self.base_field = base_field
        if title is None:
            title = base_field
        if title != "":
            wlabel = Label(base_frame, text=title, anchor=W)
            wlabel.pack(side="left", anchor=W)
            self.set_text("   ")
        

    def set_text(self, text, frame=None):
        """ Add text to current location/frame
        :text: text string to add
        :frame: frame into add default: base_frame
        """
        if frame is None:
            frame = self.base_frame
        wlabel = Label(frame, text=text, anchor=W)
        wlabel.pack(side="left", anchor=W)


    def set_sep(self, text=None, frame=None):
        """ Add default separator
        :text: text in separator
        :frame:  destination frame
        """
        if text is None:
            text = "  "
        if frame is None:
            frame = self.base_frame
        self.set_text(text, frame=frame)


    def set_vert_sep(self, frame=None, text="  "):
        """ Add default vertical separator
        :frame:  destination frame
        :text: text to place in separator
        """
        if frame is None:
            frame = self.base_frame
        sep_frame = Frame(frame)
        sep_frame.pack(side="top", anchor=N)
        self.set_text(text, frame=sep_frame)



    def set_check_box(self, frame=None, field=None,
                        label=None, value=False,
                        command=None):
        """ Set up check box for field
        :field: local field name
        :label: button label - default final section of field name
        :value: value to set "string"
        :command: function to call with new value when box changes
                    default: no call
        """
        if frame is None:
            frame = self.base_frame
        if label is None:
            label = field
            
        if label is not None:
            wlabel = Label(frame, text=label)
            wlabel.pack(side="left")
        content = BooleanVar()
        full_field = self.field_name(field)
        value = self.get_prop_val(full_field, value)
        content.set(value)
        cmd = None
        if command is not None:     # HACK - only works for ONE checkbox
            self.check_box_change_content = content
            self.check_box_change_callback = command
            cmd = self.check_box_change 
        widget =  Checkbutton(frame, variable=content, command=cmd)
        widget.pack(side="left", fill="none", expand=True)
        self.ctls[full_field] = widget
        self.ctls_vars[full_field] = content
        SlTrace.lg(f"set_check_box adding field:{full_field}")
        self.set_prop_val(full_field, value)

    def check_box_change(self):
        value = self.check_box_change_content.get()
        self.check_box_change_callback(value)


    def null_radio_call(self, val):
        SlTrace.lg(f"null_radio_call({val})")
        
    def set_radio_button(self, frame=None, field=None,
                        label=None, value=None, set_value=None,
                        command=None):
        """ Set up one radio button for field
        :field: local field name (for all buttons in this group)
        :label: button label - must be unique in this group
        :value: value associated with this button
                    default: label string
        :set_value: set group value if present
                    default: leave group_value alone
                    Anyone of the set_radio_button calls
                    can set/change the value
                    if present, value is overridden
                    by property value
        :command: function called with button group value when button selected
        
        """
        if frame is None:
            frame = self.base_frame
        if field is None:
            SelectError("set_radio_button: field is missing")
        if label is None:
            SelectError("set_radio_button - label is missing")
        if value is None:
            value = label
        if command is None:
            cmd = None
        else:
            cmd=lambda cmd=command : cmd(value)
            
            
        full_field = self.field_name(field)
        if full_field not in self.ctls_vars:
            content = StringVar()          # Create new one for group
            self.ctls_vars[full_field] = content
            ctls_ctl = self.ctls[full_field] = {}  # dictionary of Radiobutton button widgets
        else:
            content = self.ctls_vars[full_field]
            ctls_ctl = self.ctls[full_field]
        if set_value is not None:
            prop_value = self.get_prop_val(full_field, None)
            if prop_value is None:
                content.set(set_value)
                self.set_prop_val(full_field, set_value)
            else:
                content.set(prop_value)
        widget =  Radiobutton(frame, variable=content, text=label, value=value,
                               command=cmd)
        widget.pack(side="left", fill="none", expand=True)
        ctls_ctl[full_field + "." + label] = widget           # Store each button in hash

    def set_entry(self, frame=None, field=None,
                        label=None, value=None,
                        enter_command=None,
                        width=None):
        """ Set up entry
        :frame: containing frame
        :field: relative field name (after self.base_field)
        :label: field label default: no label
        :value: value to set, iff not in properties
                value's variable type is used as the entry content's type
        :enter_command:    Command to call if ENTER is keyed when
                field is in focus
                default: use self.enter_command, if present
        :returns: entry widget
        """
        if frame is None:
            frame = self.base_frame
        content = self.content_var(type(value))
        full_field = self.field_name(field)
        value = self.get_prop_val(full_field, value)
        content.set(value)
        if label is not None:
            wlabel = Label(frame, text=label)
            wlabel.pack(side="left")
            
        widget =  Entry(frame, textvariable=content, width=width)
        widget.pack(side="left", fill="none", expand=True)
        self.ctls[full_field] = widget
        self.ctls_vars[full_field] = content
        self.set_prop_val(full_field, value)
        if enter_command is not None:
            widget.bind ("<Return>", enter_command)
        elif self.enter_command is not None:
            widget.bind ("<Return>", self.enter_command)
            
        return widget

    def set_button(self, frame=None, field=None,
                        label=None, command=None):
        """ Set up button for field
        :frame: containing frame, default self.base_frame
        :field: field name
        :label: button label - default: field
        :command: command to execute when button pressed
        """
        if frame is None:
            frame = self.base_frame
        if label is None:
            label = field
            
        widget =  Button(frame, text=label, command=command)
        widget.pack(side="left", fill="none", expand=True)
        full_field = self.field_name(field)
        self.ctls[full_field] = widget
        # No variable

    def field_name(self, *fields):
        """ Create basic field name from list
        :fields: set of field segments
        """
        field_name = self.base_field
        for field in fields:
            if field_name != "":
                field_name += "."
            field_name += field
        return field_name.lower()
            
    def win_size_event(self, event):
        """ Window sizing event
        """
        win_x = self.mw.winfo_x()
        win_y = self.mw.winfo_y()
        win_width = self.mw.winfo_width()
        win_height = self.mw.winfo_height()
        self.set_window_size(win_x, win_y, win_width, win_height)

        
    def set_window_size(self, x, y, width, height, change=False):
        """ Size our window
        :change: True force window resize
        """
        self.set_prop_val("win_x", x)
        self.set_prop_val("win_y", y)
        self.set_prop_val("win_width", width)
        self.set_prop_val("win_height", height)
        if SlTrace.trace("set_window_size"):
            if ( not hasattr(self, "prev_x") or self.prev_x != x
                 or not hasattr(self, "prev_y") or self.prev_y != y
                 or not hasattr(self, "prev_width") or self.prev_width != width
                 or not hasattr(self, "prev_height") or self.prev_height != height):
                SlTrace.lg("set_window_size: change=%d x=%d y=%d width=%d height=%d" % (change, x,y,width,height))
            self.prev_x = x 
            self.prev_y = y
            self.prev_width = width
            self.prev_height = height
        if change:
            geo_str = "%dx%d+%d+%d" % (width, height, x, y)
            self.mw.geometry(geo_str)
    
    def arrange_windows(self):
        """ Arrange windows
            Get location and size for properties if any
        """
        win_x = self.get_prop_val("win_x", self.DEF_WIN_X)
        if win_x < 0:
            win_x = 50
        win_y = self.get_prop_val("win_y", self.DEF_WIN_Y)
        if win_y < 0:
            win_y = 50
        
        win_width = self.get_prop_val("win_width", self.mw.winfo_width())
        if win_width < 100:
            win_width = 100
        win_height = self.get_prop_val("win_height", self.mw.winfo_height())
        if win_height < 100:
            win_height = 100
        self.set_window_size(win_x, win_y, win_width, win_height, change=True)
        self.mw.protocol("WM_DELETE_WINDOW", self.hide_window)  # Still need info
        self.mw.bind('<Configure>', self.win_size_event)
       
    
    def get_prop_key(self, name):
        """ Translate full control name into full Properties file key
        """        
        key = self.control_prefix + "." + name
        return key

    def get_prop_val(self, name, default):
        """ Get property value as (string)
        :name: field name
        :default: default value, if not found
        :returns: "" if not found
        """
        prop_key = self.get_prop_key(name)
        prop_val = SlTrace.getProperty(prop_key)
        if prop_val is None or prop_val == "":
            return default
        if isinstance(default, bool):
            bv = str2bool(prop_val)
            return bv
        
        if isinstance(default, int):
            if prop_val == "":
                return 0
            try:
                prop_val = int(prop_val)
            except:
                try:
                    prop_val = float(prop_val)
                except:
                    prop_val = 0
                    
            return int(prop_val)
        
        if isinstance(default, float):
            if prop_val == "":
                return 0.
           
            return float(prop_val)
        else:
            return prop_val

    
    def get_val(self, name, default=None):
        """ Get current value, if any, else property value, if any,
        else default
        :name: field name
        :default: returned if not found
        """
        if name in self.vals:
            return self.vals[name]
        
        val = self.get_prop_val(name, default)
        return val


    def set_ctl(self, field_name, value):
        """ Set field, given value
        Updates field display and properties value
        :field_name: field name
        :value:        value to set
        """
        if field_name not in self.ctls_vars:
            raise SelectError("Control has no field variable %s" % field_name)
        ctl_var = self.ctls_vars[field_name]
        ctl_var.set(value)
        self.set_prop_val(field_name, value)


    def content_var(self, type):
        """ create content variable of the type val
        :type: variable type
        """
        if type == str:
            var = StringVar()
        elif type == int:
            var = IntVar()
        elif type == float:
            var = DoubleVar()
        elif type == bool:
            var = BooleanVar()
        else:
            raise SelectError(f"Unsupported content var type {type}")
        
        return var


    def set_ctl_val(self, field_name, val):
        """ Set control field
        Creates field variable if not already present
        :field_name: field name, case insensitive name
        :val: value to display
        """
        field_name = field_name.lower()
        if field_name not in self.ctls_vars:
            content_var = self.content_var(type(val))
            self.ctls_vars[field_name] = content_var
        self.ctls_vars[field_name].set(val)
 


    def set_prop_val(self, name, value):
        """ Set property value as (string)
        and internal val as value
        :name: field name
        :value: default value, if not found
        """
        self.set_val(name, value)
        prop_key = self.get_prop_key(name)
        SlTrace.setProperty(prop_key, str(value))
        
    def set_ctl_label(self, field, label):
        """ Set field's associated label, if one
        """
        if field in self.ctls_labels:
            ctl_label = self.ctls_labels[field]
            ctl_label.config(text=label)

    
    def set_val(self, name, value):
        """ Set field value
        :name: field name
        :value: value to set
        """
        self.vals[name] = value
        


    def destroy(self):
        """ Destroy window resources
        """
        SlTrace.lg("SelectControlWindow destroy")
        if self.mw is not None:
            self.mw.destroy()
        self.mw = None
        
        
    def delete_window(self):
        """ Handle window deletion
        """
        SlTrace.lg("SelectControlWindow delete_window")
        if self.play_control is not None and hasattr(self.play_control, "close_score_window"):
            self.play_control.close_score_window()
        else:
            self.destroy()
            quit()
            SlTrace.lg("Properties File: %s"% SlTrace.getPropPath())
            SlTrace.lg("Log File: %s"% SlTrace.getLogPath())
            sys.exit(0)

        self.play_control = None
        
    def hide_window(self):
        """ Hide window as we said...
        """
        self.mw.withdraw()              # Just hide, for we can't easily delete/restore a singleton


    def show_window(self):
        """ Restore to view
        """
        self.mw.deiconify()
        
if __name__ == '__main__':
        
    SlTrace.setProps()
    root = Tk()
    ###root.withdraw()       # Hide main window
    test_general = False
    test_general = True
    test_radio_button = True
    test_radio_button = False
    if test_radio_button:
        scw = SelectControlWindow(title="SelectControlWindow Testing", display=True)
        mf = Frame(root)
        mf.pack()
    
        def call_unit(unit):    
            SlTrace.lg(f"call_unit({unit})")
        
        scw.set_fields(mf, "base_field", title="distance units")
        scw.set_radio_button(frame=mf, field="second", label="meter", set_value="foot", command=call_unit)
        scw.set_radio_button(frame=mf, field="second", label="yard", command=call_unit)
        scw.set_radio_button(frame=mf, field="second", label="foot", command=call_unit)
        fname = "base_field.second"
        fval = "yard"
        SlTrace.lg(f"Settng field:{fname} to {fval}")
        scw.set_ctl_val(fname, fval)
        sf = Frame(root)
        sf.pack()
        scw.set_fields(sf, "new_field", title="RadioButton Testing")
        scw.set_radio_button(frame=sf, field="a", label="a", set_value="b")
        scw.set_radio_button(frame=sf, field="a", label="b", command=call_unit) # Only a,b calls
        scw.set_radio_button(frame=sf, field="a", label="c")
        mainloop()
    elif test_general:    
        cF = SelectControlWindow(title="SelectControlWindow Testing", display=False)
        cf2 = SelectControlWindow()
        cf2.control_display()
            
        root.mainloop()


# select_input.py    13Mar2020  crs
# from http://effbot.org/tkinterbook/tkinter-dialog-windows.htm

"""
Facilitate input / re-input with a modeless message
"""
from tkinter import Tk, Entry, Label, messagebox

from select_error import SelectError
from select_error_input import SelectErrorInput
from select_trace import SlTrace
from crs_funs import str2val

import tkSimpleDialog

class SelectInput(tkSimpleDialog.Dialog):

    def __init__(self, master=None, title=None, message=None, default=None, grab_set=False):
        """ Get input, validate format
        :master: master window,
        :title: window title
        :message: message for input
        :default: input default val/type
        :grab_set: True - grab events
        """
        self.master = master
        self.message = message
        if default is None:
            raise SelectError("Required default parameter is missing")
        
        self.default = default
        self.standalone = False
        if master is not None and hasattr(master, "tk"):
            master_focus = master
        elif master is not None and hasattr(master, "mw"):
            master_focus = master.mw
        else:
            master_focus = Tk()
            self.standalone = True
            master_focus.withdraw()               # Hide base window
        self.initial_focus = master_focus
        ###super().__init__(master_focus, title=title, grab_set=grab_set)
        super().__init__(master_focus, title=title)
        
    def body(self, master):

        Label(master, text=self.message).grid(row=0)
        Label(master, text="ENTER:").grid(row=1)
        self.e1 = Entry(master)
        self.e1.grid(row=1, column=1)
        return self.e1 # initial focus

    def validate(self):
        val_str = self.e1.get()
        try:
            self.result = str2val(val_str, self.default)
            return 1
        
        except SelectErrorInput:
            messagebox.showwarning(
                "Bad input",
                f"Illegal values {val_str}, please try again"
            )
            return 0
        
if __name__ == "__main__":
    
    root = Tk()


    si = SelectInput(root, message="SelectReport Testing", default=0)
    val = si.result
    SlTrace.lg(f"val = {val}")    



# select_control.py
"""
Adapted from block/src/select_control_window.py

Base for independent control
Provides a singleton which is universally accessible
Facilitates
    setting and retrieving of game controls
    persistent storage of values
"""
from tkinter import *

from graphics_braille.select_error import SelectError
from graphics_braille.select_trace import SlTrace


    

def str2bool(v):
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise SelectError('Not a recognized Boolean value %s' % v)
    
    
    
class SelectControl:
    CONTROL_NAME_PREFIX = "select_control"
    _instance = None         # Instance
    instance_no = 0         # Count instances
    
    def __init__(self, *args, **kwargs):
        SlTrace.lg(f"SelectControl.__init__ {SelectControl.instance_no}")
        
            
    def _init(self, play_control=None,
                control_prefix=None,
                update_report=None,
                 ):
        """ Control attributes
        :title: window title
        :update_report: function if present, called after update_settings
        """
        SelectControl.instance_no += 1
        self.play_control = play_control
        if control_prefix is None:
            control_prefix = self.CONTROL_NAME_PREFIX
        self.control_prefix = control_prefix
        self.vals = {}              # Current values if any
        self.ctls = {}              # Dictionary of field control widgets
        self.ctls_vars = {}         # Dictionary of field control widget variables
        self.ctls_textvars = {}     # Dictionary of field text variables, if requested
        self.changed = False        # Set if any value is changed vie set_val
        self.save_stack = []        # variable setting stack
        self.unsave_stack = []      # variable redo stack
        self.save_no = 0            # Track saves
        self.restore_no = 0
        self.redo_no = 0
        self.entry_no = 0           # entry position as created for display
        self.name_by_no = []      # entry names by order of creation
        self.label_no = 0           # Ordering labels, to keep name unique
        self.entry_types = {}       # Entry types: "std", "label", "internal"
        self.entry_labels = {}      # Entry labels (for display)
        self.update_report = update_report
        self.make_val("_save_no", self.save_no, entry_type="internal")
        self.make_val("_restore_no", self.restore_no, entry_type="internal")
        self.make_val("_redo_no", self.redo_no, entry_type="internal")
        
    def __new__(cls, *args, **kwargs):
        """ Make a singleton
        """
        if cls._instance is None:
            cls._instance = super(SelectControl, cls).__new__(cls)
            ###cls._instance._init(*args, **kwargs)
            cls._instance._init(*args, **kwargs)
        return cls._instance

    def add_widget(self, name, widget, repeat=False):
        """ Add control widget in sync with variables
        :name: control name
        :widget: widget
        :repeat: allow replacement of previous widget
        """
        if name in self.ctls:
            if not repeat:
                raise SelectError(f"duplicate widget add: {name}")
        
        self.ctls[name] = widget
        
    def get_val_from_ctl(self, field_name):
        """ Get value from field
        Does not set value
        :field_name: field name
        """
        field = field_name.lower()
        if field not in self.ctls_vars:
            raise SelectError("Command has no attribute %s" % field)
        value = self.ctls_vars[field].get()
        return value

    
    def set_vals(self):
        """ Read form, if displayed, and update internal values
        """
        for field in self.ctls_vars:
            self.set_val_from_ctl(field)

    def set_val_from_ctl(self, field_name):
        """ Set ctls value from field
        Also updates player value properties
        :field_name: field name
        """
        if not field_name in self.ctls_vars:
            raise SelectError("No field named %s" % field_name)
        ctl_var = self.ctls_vars[field_name]
        if ctl_var is None:
            raise SelectError("No variable for %s" % field_name)
        
        value = ctl_var.get()
        self.set_prop_val(field_name, value)
       
    
    def get_prop_key(self, name):
        """ Translate full control name into full Properties file key
        """        
        key = self.control_prefix + "." + name
        return key

    def get_prop_val(self, name, default):
        """ Get property value as (string)
        :name: field name
        :default: default value, if not found
        :returns: "" if not found
        """
        prop_key = self.get_prop_key(name)
        prop_val = SlTrace.getProperty(prop_key)
        if prop_val is None or prop_val == "":
            return default
        if isinstance(default, bool):
            bv = str2bool(prop_val)
            return bv
        
        if isinstance(default, int):
            if prop_val == "":
                return 0
            try:
                prop_val = int(prop_val)
            except:
                try:
                    prop_val = float(prop_val)
                except:
                    prop_val = 0
                    
            return int(prop_val)
        
        if isinstance(default, float):
            if prop_val == "":
                return 0.
           
            return float(prop_val)
        else:
            return prop_val

    def get_entry_type(self, name):
        """ Get entry type: std, label, internal
        :name: entry name
        """
        return self.entry_types[name]
    
    def get_val(self, name, default=None):
        """ Get current value, if any, else property value, if any,
        else default
        :name: field name
        :default: returned if not found
        """
        if name in self.vals:
            val = self.vals[name]
            if default is not None and type(val) != type(default):
                tpd = type(default)
                SlTrace.lg(f"get_val: self.vals[{name}] type{type(val)} != type{tpd} of default{default}")
                val = tpd(val)
            elif type(val) == str and re.match(r'^\d+$', val):
                SlTrace.lg(f"get_val: suspect {name} {val} is an int")
                return int(val)
            elif type(val) == str and re.match(r'^(\d*)?\.(\d*)$', val) and val != ".":
                SlTrace.lg(f"get_val: suspect {name} {val} is a float")
                return float(val)
                    
            return val
        
        val = self.get_prop_val(name, default)
        return val

    def get_textvar(self, name):
        """ Get variable for field
        :name: field name
        :returns: field control variable
        """
        if name not in self.ctls_textvars:
            raise SelectError(f"text field {name} not present")
        return self.ctls_textvars[name]

    def get_var(self, name):
        """ Get variable for field
        :name: field name
        :returns: field control variable
        """
        if name not in self.ctls_vars:
            raise SelectError(f"field {name} not present")
        return self.ctls_vars[name]

    def get_var_label(self, name):
        """ Get variable label for field
        :name: field name
        :returns: field label
        """
        if name not in self.ctls_vars:
            raise SelectError(f"field {name} not present")
        return self.entry_labels[name]

    def get_var_names(self, std=True, internal=False, label=False):
        """ Get list of variable names
        in order of creation
        :std: True include standard names
                default: True
        :internal:  Include special entry names 
                _... internal values
                default: False
        :label: Include group labels
                default: False
        """
        names = []
        for ie in range(len(self.name_by_no)):
            name = self.name_by_no[ie]
            entry_type = self.entry_types[name]
            if entry_type == "internal":
                if internal:
                    names.append(name)
            elif entry_type == "std":
                if std:
                    names.append(name)
            elif entry_type == "label":
                if label:
                    names.append(name)
                
        return names
        
    def list_vals(self, prefix=None):
        """ List all key,value pairs
        :prefix: Optional text prefix
        """
        if prefix is None:
            prefix = ""
        for name in self.vals:
            entry_type = self.get_entry_type(name)
            val = self.vals[name]
            SlTrace.lg(f"{prefix} {name}({entry_type}): {val}")
    
    def list_textvals(self, prefix=None):
        """ List all key,value text variable pairs
        :prefix: Optional text prefix
        """
        if prefix is None:
            prefix = ""
        for name in self.ctls_textvars:
            var = self.ctls_textvars[name]
            SlTrace.lg(f"{prefix} {name}: {var.get()}")
            

    def make_label(self, label):
        """ Make label for control panel display
        :label: label string
        """
        self.label_no += 1
        name = f"label_{self.label_no}"
        self.make_val(name, default=label, entry_type="label", label=label)
        
    def make_val(self, name, default=None, repeat=False, textvar=False,
                 entry_type=None, label=None):
        """ Create field, if not present, returning value
        :name: field name
        :default: value used if field not present
        :repeat: True->allow repeats default: False
        :textvar: True - create text field variable for widjets such as Entry etc.
        :entry_type: entry type: "std" standard variable
                                "label" display label
                                "internal" internal variable
        :label: Entry label text, default: use name
        :returns: value set
        """
        if default is None:
            raise SelectError(f"make_val[{name}] REQUIRED default is Missing")
        
        if name in self.vals:
            if not repeat:
                raise SelectError(f"name: {name} already made")
        if label is None:
            label = name
        self.entry_labels[name] = label
                
        if entry_type is None:
            entry_type = "std"
        
        if name not in self.vals:
            self.entry_types[name] = entry_type
            self.name_by_no.append(name)
            self.entry_no = len(self.name_by_no)
            self.ctls_vars[name] = self.content_var(type(default))
            self.vals[name] = default
        prop_key = self.get_prop_key(name)
        prop_val = SlTrace.getProperty(prop_key, default=None)
        if prop_val is None:
            val = default
        else:
            tpd = type(default)
            if tpd is str:
                val = prop_val            # Keep the string
            elif tpd is bool:
                val = str2bool(prop_val)
            else:
                val = tpd(prop_val)         # Just cast it
        self.set_val(name, val)
        if textvar:
            self.ctls_textvars[name] = StringVar()
            self.ctls_textvars[name].set(val)
        return val
    
    def set_ctl(self, field_name, value):
        """ Set field, given value
        Updates field display and properties value
        :field_name: field name
        :value:        value to set
        """
        if field_name not in self.ctls_vars:
            raise SelectError(f"Control has no field variable {field_name}")
        ctl_var = self.ctls_vars[field_name]
        ctl_var.set(value)
        self.set_prop_val(field_name, value)


    def content_var(self, ctype):
        """ create content variable of the ctype val
        :ctype: variable type
        """
        if ctype == str:
            var = StringVar()
        elif ctype == int:
            var = IntVar()
        elif ctype == float:
            var = DoubleVar()
        elif ctype == bool:
            var = BooleanVar()
        else:
            raise SelectError(f"Unsupported content var type {ctype}")
        
        return var


    def set_ctl_val(self, field_name, val):
        """ Set control field
        Creates field variable if not already present
        Sets vals[field_name]
        :field_name: field name
        :val: value to display
        """
        if field_name not in self.ctls_vars:
            content_var = self.content_var(type(val))
            self.ctls_vars[field_name] = content_var
        self.ctls_vars[field_name].set(val)
        self.vars[field_name] = val


    def set_prop_val(self, name, value):
        """ Set local (value) and property value(string)
        :name: field name
        :value: default value, if not found
        """
        self.vals[name] = value         # set current value
        prop_key = self.get_prop_key(name)
        SlTrace.setProperty(prop_key, str(value))

    def update_prop_val(self, name):
        """ Update properties value from current value
        :name: field name
        """
        val = self.vals[name]
        self.set_prop_val(name, val)

    def update_settings(self, force=False):
        """ Update all control vars properties
            Save previous settings if any changes or force is True
                if text variables, use widget contents
                    to update textvar, and var, and prop
                else use control vars to update prop
            :force: Always save settings default: False
        """
        trace_str = ""
                                    # Check if any changes
        for name in self.ctls_vars:
            is_changed = False
            var = self.ctls_vars[name]
            val = var.get()
            if name in self.ctls_textvars and name in self.ctls:
                widget = self.ctls[name]
                new_val = widget.get()
            else:
                var = self.ctls_vars[name]
                new_val = var.get()
            if new_val != val:
                is_changed = True
            if is_changed and SlTrace.trace("track_vals"):
                if trace_str != "":
                    trace_str += " "
                val_str = f"was: {val} is {new_val}"
                trace_str += f"Change: {name}:{val_str}"
        if force or is_changed:
            self.save_settings()
            
        for name in self.ctls_vars:
            if name in self.ctls_textvars and name in self.ctls:
                widget = self.ctls[name]
                val = widget.get()
                self.set_val(name, val)
            else:
                var = self.ctls_vars[name]
                val = var.get()
                self.set_val(name, val)
            if SlTrace.trace("track_vals"):
                if trace_str != "":
                    trace_str += " "
                val_str = self.get_prop_val(name, "???")
                trace_str += f"{name}:{val_str}"
        if trace_str != "":
            SlTrace.lg(f"update_settings: {trace_str}")
        if self.update_report is not None:
            self.update_report(self)

    def restore_vals(self, redo=False):
        """ Backup current values, clearing changed flag
        :redo: if True, restore before last undo
        :returns: True if successful else False
        """
        
        restored_entry = None
        self.restore_no += 1
        SlTrace.lg(f"restore_vals restore_no: {self.restore_no}", "track_vals")
        self.set_val("_restore_no", self.restore_no, no_change=True)
        if redo:
            if len(self.unsave_stack)  > 0:
                restored_entry = self.unsave_stack.pop()
        else:
            if len(self.save_stack) > 0:
                restored_entry = self.save_stack.pop()
        if restored_entry is not None:
            cur_vals = self.get_settings()
            self.save_vals(entry=cur_vals, to_redo=not redo)
            for name in restored_entry:
                self.set_val(name, restored_entry[name])
            self.changed = False
            SlTrace.lg("restore_vals")
            se = restored_entry
            for name in se:
                SlTrace.lg(f"{name}: {se[name]}")
            return True
        
        else:
            SlTrace.lg(f"empty {'redo' if redo else 'save'} stack")
            return False
        

    def save_if_change(self):
        """ Save values if any change since last save_vals
        """
        if self.changed:
            self.save_vals()


    def save_settings(self):
        """ Save current saveings
        """
        self.save_vals()

    def get_settings(self):
        """ Get current value settings
        :returns: current value settings
        """
        entry = {}
        for name in self.ctls_vars:
            entry[name] = self.get_val(name)
        return entry    
                    
    def save_vals(self, entry=None, to_redo=False):
        """ Backup current values, or others, clearing changed flag
        :entry: values dictionary defalut: use self.ctls_vars
        :to_redo: if True save to unsave_stack
        """
        if entry is None:
            entry = self.get_settings()            
        self.save_no += 1
        SlTrace.lg(f"save_vals save_no: {self.save_no}", "track_vals")
        self.set_val("_save_no", self.save_no, no_change=True)
        if to_redo:
            self.unsave_stack.append(entry)
        else:
            self.save_stack.append(entry)
        self.changed = False
        SlTrace.lg("save_vals", "track_vals")
        se = entry
        for name in se:
            SlTrace.lg(f"{name}: {se[name]}", "track_vals")
            
    def undo_settings(self):
        """ Recover previous saved window settings
        """
        self.restore_vals()
    
    def redo_settings(self):
        """ Recover settings replaced by last undo
        """
        self.redo_no += 1
        SlTrace.lg(f"redo_settings redo_no: {self.redo_no}", "track_vals")
        self.redo_no += 1
        self.set_val("_redo_no", self.redo_no, no_change=True)
        self.restore_vals(redo=True)
                        
    def set_val(self, name, value, no_change=False):
        """ Set field value, updating property value
        :name: field name
        :value: value to set
        :no_change: True -> ignore change default:False
        :returns: value to facilitate compact variable updating
        """
        tvalue = type(value)
        tval = type(self.vals[name])
        if tvalue != tval:
            if tvalue == str and tval == bool:
                value = str2bool(value)
            else:
                if tvalue == str:
                    if value == "":
                        value = "0"
                value = tval(value)
                    
        if value != self.vals[name]:
            if not no_change:
                self.changed = True
        self.vals[name] = value           # Update properties
        self.update_prop_val(name)
        if name in self.ctls_textvars:
            self.ctls_textvars[name].set(value)
        if name in self.ctls_vars:
            self.ctls_vars[name].set(value)
        return value
    
if __name__ == '__main__':
    from tkinter import *       # IntVar() fails without
    import argparse
    
    SlTrace.setProps()
    
    
    mw = Tk()
    cF = SelectControl(control_prefix="select_control_self_test")
    cF.make_label("Puzzle Dimensions")
    ncol = cF.make_val("ncol", 4)     # default, override from properties
    nrow = cF.make_val("nrow", 5)
    size = cF.make_val("size", 123.456, textvar=True)
    SlTrace.lg(f"After make_val ncol={ncol} nrow={nrow} size={size}")    
    cF.list_vals("Pre-command line parsing")
    pcl_ncol = ncol
    pcl_nrow = nrow
    pcl_size = size
    SlTrace.lg(f"Initializing values ncol={ncol} nrow={nrow} size={size}")    
    cF.update_settings()
    
    parser = argparse.ArgumentParser()
    parser.add_argument('--ncol=', type=int, dest='ncol', default=ncol)
    parser.add_argument('--nrow=', type=int, dest='nrow', default=nrow)
    args = parser.parse_args()             # or die "Illegal options"
    SlTrace.lg(f"args: {args}\n")
    ncol = cF.set_val("ncol", args.ncol)
    nrow = cF.set_val("nrow", args.nrow)
    SlTrace.lg(f"After command line parsing ncol={ncol} nrow={nrow} size={size}")
    cF.update_settings()    
    cF.list_vals("starting vals:")
    cF.update_settings()
    ncol = 2
    nrow = 3
    size = 789.
    cF.set_val("ncol", ncol)
    cF.set_val("nrow", nrow)
    cF.set_val("size", size)
    cF.list_vals(f"After ncol={cF.get_val('ncol')} nrow={cF.get_val('nrow')} size={cF.get_val('size')}")
    cF.update_settings()
    cF.update_settings()     # No changes should be seen/saved
    
    size2 = size + .1
    cF.make_val("size", size2, repeat=True, textvar=True)
    ncol = 8
    nrow = 8
    size = 400.
    cF.set_val("ncol", ncol)
    cF.set_val("nrow", nrow)
    cF.set_val("size", size)
    SlTrace.lg(f"After ncol={cF.get_val('ncol')} nrow={cF.get_val('nrow')} size={cF.get_val('size')}")
    cF.list_vals()
    cF.list_textvals("textvars")
    
    cF.undo_settings()
    SlTrace.lg("After undo_settings")
    cF.list_vals()
    cF.list_textvals("textvars")
    
    cF.redo_settings()
    SlTrace.lg("After redo_settings")
    cF.list_vals()
    cF.list_textvals("textvars")
    
    cF.set_val("ncol", pcl_ncol+1)
    cF.set_val("nrow", pcl_nrow+1)
    cF.set_val("size", pcl_size+1)
    SlTrace.lg(f"After ncol={cF.get_val('ncol')} nrow={cF.get_val('nrow')} size={cF.get_val('size')}")
    cF.list_vals()
    cF.list_textvals("textvars")
    cF.update_settings()
    
    cF.set_val("ncol", pcl_ncol+2)
    cF.set_val("nrow", pcl_nrow+2)
    cF.set_val("size", pcl_size+3)
    SlTrace.lg(f"After ncol={cF.get_val('ncol')} nrow={cF.get_val('nrow')} size={cF.get_val('size')}")
    cF.list_vals()
    cF.list_textvals("textvars")


    cF.undo_settings()
    SlTrace.lg(f"After undo_settings")
    cF.list_vals()
    cF.list_textvals("textvals")

    cF.undo_settings()
    SlTrace.lg(f"After undo_settings")
    cF.list_vals()
    cF.list_textvals("textvals")

    cF.undo_settings()
    SlTrace.lg(f"After undo_settings")
    cF.list_vals()
    cF.list_textvals("textvals")

    cF.redo_settings()
    SlTrace.lg(f"After undo_settings redo=True")
    cF.list_vals()
    cF.list_textvals("textvais")

    cF.redo_settings()
    SlTrace.lg(f"After undo_settings redo=True")
    cF.list_vals()
    cF.list_textvals("textvals")



#select_window.py 19Sep2018
"""
Program Level Menu control
 From PoolWindow
"""
import sys
import os
from tkinter import *
from select_trace import SlTrace
from trace_control_window import TraceControlWindow
from arrange_control import ArrangeControl

# Here, we are creating our class, Window, and inheriting from the Frame
# class. Frame is a class from the tkinter module. (see Lib/tkinter/__init__)
class SelectWindow(Frame):
    CONTROL_NAME_PREFIX = "play_control"

    def __deepcopy__(self, memo=None):
        """ provide deep copy by just passing shallow copy of self,
        avoiding tkparts inside sel_area
        """
        SlTrace.lg("SelectWindow __deepcopy__", "copy")
        return self
            
        
    # Define settings upon initialization. Here you can specify
    ###@profile    
    def __init__(self,
                 master=None,
                 title=None,
                 control_prefix=None,
                 pgmExit=None,
                 file_open=None,
                 file_save=None,                 
                 cmd_proc=False,
                 cmd_file=None,
                 arrange=None,
                 arrange_selection=False,
                 game_control=None,
                 games=[],          # text, proc pairs
                 actions=[],
                 ):
        """ Setup window controls
        :arrange_selection: - incude arrangement controls
                        default: False
        """
        # parameters that you want to send through the Frame class. 
        Frame.__init__(self, master)   

        #reference to the master widget, which is the tk window                 
        self.title = title
        self.master = master
        self.file_open = file_open
        self.file_save = file_save
        if control_prefix is None:
            control_prefix = SelectWindow.CONTROL_NAME_PREFIX
        self.control_prefix = control_prefix
        self.arrange = arrange
        if arrange is not None:
            arrange_selection = False
        self.arrange_selection = arrange_selection
        self.pgmExit = pgmExit
        self.game_control = game_control
        master.protocol("WM_DELETE_WINDOW", self.pgm_exit)
        self.games = games
        self.actions = actions
        self.tc = None          # Trace control
        self.arc = None         # Arrangement control
        self.arc_call_d = {}     # arc call back functions
        self.cmd_proc = cmd_proc    # Setup command file processing
        self.cmd_file = cmd_file    # if not None, execute this cmd file
        #with that, we want to then run init_window, which doesn't yet exist
        self.init_window()

        
    #Creation of init_window
    def init_window(self):

        # changing the title of our master widget 
        if self.title is not None:
            self.master.title(self.title)

        # allowing the widget to take the full space of the root window
        self.pack(fill=BOTH, expand=False)

        # creating a menu instance
        menubar = Menu(self.master)
        self.menubar = menubar      # Save for future reference
        self.master.config(menu=menubar)

        # create the file object)
        file_open_cmd = self.file_open
        if file_open_cmd is None:
            file_open_cmd = self.File_Open_tbd
        filemenu = Menu(menubar, tearoff=0)
        filemenu.add_command(label="Open", command=file_open_cmd)
        file_save_cmd = self.file_save
        if file_save_cmd is None:
            file_save_cmd = self.File_Save_tbd
        filemenu.add_command(label="Save", command=file_save_cmd)
        filemenu.add_separator()
        filemenu.add_command(label="Log", command=self.LogFile)
        filemenu.add_command(label="Properties", command=self.Properties)
        filemenu.add_separator()
        ###filemenu.add_comand(label="Cmd", command=self.command_proc)
        filemenu.add_separator()
        filemenu.add_command(label="Exit", command=self.pgmExit)
        menubar.add_cascade(label="File", menu=filemenu)
        if self.arrange is not None:         # Explicit arrange control cmd
            menubar.add_command(label="Arrange",
                                 command=self.arrange)
                            
        elif self.arrange_selection:    # Arrange control - optoinal
            menubar.add_command(label="Arrange",
                                 command=self.arrange_control)

                                # Trace control
        menubar.add_command(label="Trace", command=self.trace_control)
        self.arrange_windows()
        self.master.bind( '<Configure>', self.win_size_event)


    def get_game_control(self):
        """ Retrieve game control to pass to SelectPlay
        """
        return self.game_control
    

    def win_size_event(self, event):
        """ Window sizing event
        """
        win_x = self.master.winfo_x()
        win_y = self.master.winfo_y()
        win_width = self.master.winfo_width()
        win_height = self.master.winfo_height()
        self.set_prop_val("win_x", win_x)
        self.set_prop_val("win_y", win_x)
        self.set_prop_val("win_width", win_width)
        self.set_prop_val("win_height", win_height)
    
    def arrange_windows(self):
        """ Arrange windows
            Get location and size for properties if any
        """
        win_x = self.get_prop_val("win_x", 50)
        if win_x < 0 or win_x > 1400:
            win_x = 50
        win_y = self.get_prop_val("win_y", 50)
        if win_y < 0 or win_y > 1400:
            win_y = 50
        
        win_width = self.get_prop_val("win_width", self.master.winfo_width())
        win_height = self.get_prop_val("win_height", self.master.winfo_height())
        geo_str = "%dx%d+%d+%d" % (win_width, win_height, win_x, win_y)
        self.master.geometry(geo_str)
        
    def geometry(self, geo):
        """ Adjust parent window size/location
        """
        self.master.geometry(geo)
    
    def get_prop_key(self, name):
        """ Translate full  control name into full Properties file key
        """        
        key = self.control_prefix + "." + name
        return key

    def get_prop_val(self, name, default):
        """ Get property value as (string)
        :name: field name
        :default: default value, if not found
        :returns: "" if not found
        """
        prop_key = self.get_prop_key(name)
        prop_val = SlTrace.getProperty(prop_key)
        if prop_val is None:
            return default
        
        if isinstance(default, int):
            if prop_val == "":
                return 0
           
            return int(prop_val)
        elif isinstance(default, float):
            if prop_val == "":
                return 0.
           
            return float(prop_val)
        else:
            return prop_val

    def set_prop_val(self, name, value):
        """ Set property value as (string)
        :name: field name
        :value: default value, if not found
        """
        prop_key = self.get_prop_key(name)
        SlTrace.setProperty(prop_key, str(value))
        
        
    def pgm_exit(self):
        if self.pgmExit is not None:
            self.pgmExit()
        else:
            sys.exit()    
            
    def File_Open_tbd(self):
        print("File_Open_menu to be determined")

    def File_Save_tbd(self):
        print("File_Save_menu to be determined")

    def add_menu(self):
        """ Add pull-down menu to top menu bar
        :label: pull-down label
        :returns: menubar, filemenu to which one can
            filemenu.add_command(label="Log", command=self.LogFile)
            filemenu.add_separator()
            ...
            menubar.add_cascade(label="menu-title", menu=filemenu)
           
        """
        filemenu = Menu(self.menubar, tearoff=0)
        return self.menubar, filemenu
 
    def add_menu_command(self, label=None, call_back=None):
        """ Add simple menu command to top menu
        :label: command label
        :call_back: function to be called when selected
        """
        self.menubar.add_command(label=label, command=call_back)

    def add_menu_separator(self):
        """ Add simple menu separator to top menu
        """
        self.menubar.add_separator()


    def command_proc(self):
        """ Setup command processing options / action
        """
        
        
    def get_arc(self):
        """ Return reference to arrange control
        """
        return self.arc
    
    
    def LogFile(self):
        print("Display Log File")
        abs_logName = SlTrace.getLogPath()
        SlTrace.lg("Log file  %s"
                    % abs_logName)
        ###osCommandString = "notepad.exe %s" % abs_propName
        ###os.system(osCommandString)
        import subprocess as sp
        programName = "notepad.exe"
        sp.Popen([programName, abs_logName])
    
    
    def Properties(self):
        print("Display Properties File")
        abs_propName = SlTrace.defaultProps.get_path()
        SlTrace.lg("properties file  %s"
                    % abs_propName)
        ###osCommandString = "notepad.exe %s" % abs_propName
        ###os.system(osCommandString)
        import subprocess as sp
        programName = "notepad.exe"
        sp.Popen([programName, abs_propName])
        
        
    def select_all(self):
        if self.tc is None:
            self.select_trace()
        self.tc.select_all()
        
            
    def select_none(self):
        if self.tc is None:
            self.select_trace()
        self.tc.select_none()

        

    def arrange_control(self):
        """ Create arrangement window
        :returns: ref to ArrangeControl object
        """
        if self.arc is not None:
            self.arc.delete_window()
            self.arc = None
        
        self.arc = ArrangeControl(self, title="Arrange")
        for callname, callfn in self.arc_call_d.items():     # Enable any call back functions
            self.arc.set_call(callname, callfn)
        return self.arc

    def ctl_list(self, ctl_name, selection_list):
        return self.arc.ctl_list(ctl_name, selection_list)


    def get_ctl_entry(self, name):
        """ Get control value.  If none return default
        """
        if self.arc is None:
            return None
        return self.arc.get_entry_val(name)
 

    def get_current_val(self, name, default=None):
        """ Get control value.  If none return default
        """
        if self.arc is None:
            return default
        return self.arc.get_current_val(name, default)


    def get_component_val(self, name, comp_name, default=None):
        """ Get component value of named control
        Get value from widget, if present, else use entry value
        """
        if self.arc is None:
            return default
        return self.arc.get_component_val(name, comp_name, default)

    def get_component_next_val(self, base_name,
                            nrange=50,
                            inc_dir=1,
                            default_value=None):
        """ Next value for this component
        :control_name: control name
        :comp_name: component name
        :nrange: - number of samples for incremental
        :default_value: default value
        """
        return self.arc.get_component_next_val(base_name,
                                nrange=nrange, inc_dir=inc_dir, default_value=default_value)


    def get_inc_val(self, name, default):
        """ Get inc value.  If none return default
        """
        if self.arc is None:
            return default
        return self.arc.get_inc_val(name, default)
 

    def set_current_val(self, name, val):
        """ Set current value.
        """
        if self.arc is None:
            return
        return self.arc.set_current_val(name, val)
 

    def set_component_val(self, name, comp_name, val):
        """ Set current value.
        """
        if self.arc is None:
            return
        return self.arc.set_component_val(name, comp_name, val)
        
    
    def set_call(self, name, function):
        """ Set for call back from arrange control
           1. If arc present via arc
           2. Else store for later enabling when arc is created
        """
        if self.arc is not None:
            self.arc.set_call(name, function)
        else:
            self.arc_call_d[name] = function
 
        

    def trace_control(self):
 
        def report_change(flag, val, cklist=None):
            SlTrace.lg("changed: %s = %d" % (flag, val), "controls")
            new_val = SlTrace.getLevel(flag)
            SlTrace.lg("New val: %s = %d" % (flag, new_val), "controls")
            if cklist is not None:
                cklist.list_ckbuttons()
        
        if self.tc is not None:
            self.tc.delete_tc_window()
            self.tc = None
        
        self.tc = TraceControlWindow(self, change_call=report_change)


    def tc_destroy(self):
        """ Called if TraceControlWindow window closes
        """
        self.tc = None


    def update_form(self):
        """ Update any field changes
        """
        if self.arc is not None:
            self.arc.update_form()
#########################################################################
#          Self Test                                                    #
#########################################################################
if __name__ == "__main__":
    from trace_control_window import TraceControlWindow    
        
    # root window created. Here, that would be the only window, but
    # you can later have windows within windows.
    mw = Tk()
    SlTrace.lg("Startup")
    
    def user_exit():
        print("user_exit")
        print("Calling SlTrace.onexit()")
        SlTrace.onexit()
        exit()
    
    def user_file_open():
        print("user_file_open")
    
    def user_file_save():
        print("user_file_save")
            
            
    SlTrace.setProps()
    set_flags = True
    set_flags = False
    if set_flags:
        SlTrace.lg("setFlags")
        SlTrace.setFlags("flag1=1,flag2=0,flag3=1,flag4=0, flag5=1, flag6=1")
    else:
        SlTrace.lg("no setFlags")
        
    mw.geometry("400x300")
    
    #creation of an instance
    app = SelectWindow(mw,
                    title="select_window Testing",
                    pgmExit=user_exit,
                    arrange_selection=False,
                    file_open = user_file_open,
                    file_save = user_file_save,
                    )

    show_trace = True
    show_trace = False
    if show_trace:
        app.trace_control()

    
    #mainloop 
    mw.mainloop()  




#onedrive_ck.py 19Nov2025  from google search

import os
from glob import glob

def is_desktop_on_onedrive():
    """
    Determines if the Desktop folder is located within the OneDrive directory.
    """
    try:
        # Attempt to find the Desktop path using glob, which can handle OneDrive's relocated Desktop
        # This looks for a "Desktop" folder within any subfolder of the user's home directory.
        desktop_path_glob_list = glob(os.path.expanduser("~\\*\\Desktop"))

        if desktop_path_glob_list:
            # If glob found a path, check if it contains "OneDrive"
            # This handles cases where Desktop is directly under a OneDrive folder
            for path in desktop_path_glob_list:
                if "OneDrive" in path:
                    return True, path
            
            # If glob found a path but it's not directly in "OneDrive",
            # it might be a standard Desktop path or a different redirection.
            # We can then compare it to the "default" Desktop location.
            default_desktop_path = os.path.expanduser("~\\Desktop")
            if default_desktop_path not in desktop_path_glob_list:
                # If the glob path is different from the default, and not directly OneDrive,
                # it's still possible it's a OneDrive-managed path, but we need more robust checks.
                # For simplicity here, we'll assume a direct "OneDrive" check is sufficient for this method.
                pass

        # If glob returns an empty list, it means the Desktop is likely in the default location
        # or a location not covered by the glob pattern.
        # In this case, we check the standard Desktop path for "OneDrive".
        standard_desktop_path = os.path.expanduser("~\\Desktop")
        if "OneDrive" in standard_desktop_path:
            return True, standard_desktop_path
            
        return False, standard_desktop_path # Return False and the standard path if not found in OneDrive

    except Exception as e:
        print(f"An error occurred: {e}")
        return False, None

if __name__ == "__main__":
    on_onedrive, desktop_location = is_desktop_on_onedrive()
    if on_onedrive:
        print(f"The Desktop is managed by OneDrive. Location: {desktop_location}")
    else:
        print(f"The Desktop is not managed by OneDrive. Location: {desktop_location}")



# select_dialog.py    13Mar2020  crs from http://effbot.org/tkinterbook/tkinter-dialog-windows.htm


import tkSimpleDialog

class SelectDialog(tkSimpleDialog.Dialog):

    def body(self, master):

        Label(master, text="First:").grid(row=0)
        Label(master, text="Second:").grid(row=1)

        self.e1 = Entry(master)
        self.e2 = Entry(master)

        self.e1.grid(row=0, column=1)
        self.e2.grid(row=1, column=1)
        return self.e1 # initial focus

    def apply(self):
        first = int(self.e1.get())
        second = int(self.e2.get())
        print(f"first={first}, second={second}")

if __name__ == "__main__":
    from tkinter import *

    root = Tk()


    d = SelectDialog(root)
    print(f"d={d.result}")


#robust_queue_test.py

from multiprocessing import Process
from time import sleep
from random import randint

import graphics_braille
import robust_queue as rq

# A simple use case of the custom Queue that allows .qsize() method
# in MacOS X.

def foo(q):
    i = 0
    while True:
        q.put(f'current i = {i}')
        sleep(randint(0, 3))
        i += 1


if __name__ == '__main__':
    q: rq.RQueue = rq.RQueue()
    p: Process = Process(target=foo, args=(q,))
    p.start()

    times = 0
    while times < 5:
        print(f'current qsize = {q.qsize()}')
        if not q.empty():
            print(f'qsize = {q.qsize()} before get')
            print(f'Item got from queue: {q.get()}')
            print(f'qsize = {q.qsize()} after get')
        times += 1
        sleep(randint(0, 3))

    p.terminate()
    p.join()
    print(f'qsize = {q.qsize()} at the end')


#robust_queue.py

from multiprocessing.queues import Queue
import multiprocessing

# The following implementation of custom RQueue to avoid NotImplementedError
# when calling queue.qsize() in MacOS X comes almost entirely from this github
# discussion: https://github.com/keras-team/autokeras/issues/368
# Necessary modification is made to make the code compatible with Python3.


class SharedCounter(object):
    """ A synchronized shared counter.
    The locking done by multiprocessing.Value ensures that only a single
    process or thread may read or write the in-memory ctypes object. However,
    in order to do n += 1, Python performs a read followed by a write, so a
    second process may read the old value before the new one is written by the
    first process. The solution is to use a multiprocessing.Lock to guarantee
    the atomicity of the modifications to Value.
    This class comes almost entirely from Eli Bendersky's blog:
    http://eli.thegreenplace.net/2012/01/04/shared-counter-with-pythons-multiprocessing/
    """

    def __init__(self, n=0):
        self.count = multiprocessing.Value('i', n)

    def increment(self, n=1):
        """ Increment the counter by n (default = 1) """
        with self.count.get_lock():
            self.count.value += n

    @property
    def value(self):
        """ Return the value of the counter """
        return self.count.value


class RQueue(Queue):
    """ A portable implementation of multiprocessing.Queue.
    Because of multithreading / multiprocessing semantics, Queue.qsize() may
    raise the NotImplementedError exception on Unix platforms like Mac OS X
    where sem_getvalue() is not implemented. This subclass addresses this
    problem by using a synchronized shared counter (initialized to zero) and
    increasing / decreasing its value every time the put() and get() methods
    are called, respectively. This not only prevents NotImplementedError from
    being raised, but also allows us to implement a reliable version of both
    qsize() and empty().

    Note the implementation of __getstate__ and __setstate__ which help to
    serialize RQueue when it is passed between processes. If these functions
    are not defined, RQueue cannot be serialized, which will lead to the error
    of "AttributeError: 'RQueue' object has no attribute 'size'".

    See the answer provided here: https://stackoverflow.com/a/65513291/9723036
    
    For documentation of using __getstate__ and __setstate__ to serialize objects,
    refer to here: https://docs.python.org/3/library/pickle.html#pickling-class-instances
    """

    def __init__(self, size=None):
        """ Setup r obust queue
        :size: max queue size - NOT implemented/used
        """
        self.max_size = size
        super().__init__(ctx=multiprocessing.get_context())
        self.size = SharedCounter(0)

    def __getstate__(self):
        """Help to make RQueue instance serializable.

        Note that we record the parent class state, which is the state of the
        actual queue, and the size of the queue, which is the state of RQueue.

        self.size is a SharedCounter instance. It is itself serializable.
        """
        return {
            'parent_state': super().__getstate__(),
            'size': self.size,
        }

    def __setstate__(self, state):
        super().__setstate__(state['parent_state'])
        self.size = state['size']

    def put(self, *args, **kwargs):
        super().put(*args, **kwargs)
        self.size.increment(1)

    def get(self, *args, **kwargs):
        item = super().get(*args, **kwargs)
        self.size.increment(-1)
        return item

    def qsize(self):
        """ Reliable implementation of multiprocessing.Queue.qsize() """
        return self.size.value

    def empty(self):
        """ Reliable implementation of multiprocessing.Queue.empty() """
        return not self.qsize()